__version__ = '1.0.5'

from . import base, primary, secondary
from .base import (get_time, get_time_monthly_vol,
                   DeclineCurve, PrimaryPhase, SecondaryPhase,
                   DAYS_PER_MONTH, DAYS_PER_YEAR)
from .primary import NullPrimaryPhase, MultisegmentHyperbolic, MH, THM, PLE, SE, Duong
from .secondary import NullSecondaryPhase, PLYield
from .bourdet import bourdet



"""
Decline Curve Models
Copyright © 2020 David S. Fulford

Author
------
David S. Fulford
Derrick W. Turk

Notes
-----
Created on August 5, 2019
"""

import warnings

import dataclasses as dc
from dataclasses import dataclass

from numpy import ndarray
import numpy as np

from scipy.integrate import fixed_quad  # type: ignore

from typing import (TypeVar, Type, List, Dict, Tuple, Any,
                    Sequence, Optional, Callable, ClassVar, Union)
from typing import cast

from .base import (ParamDesc, DeclineCurve, PrimaryPhase, SecondaryPhase,
                   DAYS_PER_MONTH, DAYS_PER_YEAR)


@dataclass
class NullSecondaryPhase(SecondaryPhase):
    """
    A null `SecondaryPhase` class that always returns zeroes.

    Parameters
    ----------
      None
    """

    def _set_defaults(self) -> None:
        # Do not associate with the null secondary phase
        pass

    def _yieldfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _qfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Nfn(self, t: ndarray, **kwargs: Dict[Any, Any]) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Dfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Dfn2(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _betafn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _bfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return []


@dataclass(frozen=True)
class PLYield(SecondaryPhase):
    """
    Power-Law Secondary Phase Model.

    Fulford, D.S. 2018. A Model-Based Diagnostic Workflow for Time-Rate
    Performance of Unconventional Wells. Presented at Unconventional Resources
    Conference in Houston, Texas, USA, 23–25 July. URTeC-2903036.
    https://doi.org/10.15530/urtec-2018-2903036.

    Has the general form of

    .. math::

        GOR = c \, t^m

    and allows independent early-time and late-time slopes ``m0`` and ``m`` respectively.

    Parameters
    ----------
      c: float
        The value of GOR that acts as the anchor or pivot at ``t=t0``.

      m0: float
        Early-time power-law slope.

      m: float
        Late-time power-law slope.

      t0: float
        The time of the anchor or pivot value ``c``.

      min: Optional[float] = None
        The minimum allowed value. Would be used e.g. to limit minimum CGR.

      max: Optional[float] = None
        The maximum allowed value. Would be used e.g. to limit maximum GOR.
    """
    c: float
    m0: float
    m: float
    t0: float
    min: Optional[float] = None
    max: Optional[float] = None

    def _validate(self) -> None:
        pass

    def _yieldfn(self, t: ndarray) -> ndarray:
        c = self.c
        t0 = self.t0
        m = np.where(t < t0, self.m0, self.m)

        with warnings.catch_warnings(record=True) as w:
            if self.min is not None or self.max is not None:
                return np.where(t == 0.0, 0.0, np.clip(c * (t / t0) ** m, self.min, self.max))
            return np.where(t == 0.0, 0.0, c * (t / t0) ** m)

    def _qfn(self, t: ndarray) -> ndarray:
        return self._yieldfn(t) / 1000.0 * self.primary._qfn(t)

    def _Nfn(self, t: ndarray, **kwargs: Dict[Any, Any]) -> ndarray:
        with warnings.catch_warnings(record=True) as w:
            return self._integrate_with(self._qfn, t, **kwargs)

    def _Dfn(self, t: ndarray) -> ndarray:
        c = self.c
        t0 = self.t0
        m = np.where(t < t0, self.m0, self.m)
        y = self._yieldfn(t)

        if self.min is not None:
            m[y <= self.min] = 0.0
        if self.max is not None:
            m[y >= self.max] = 0.0
        return -m / t + self.primary._Dfn(t)

    def _Dfn2(self, t: ndarray) -> ndarray:
        c = self.c
        t0 = self.t0
        m = np.where(t < t0, self.m0, self.m)
        y = self._yieldfn(t)

        if self.min is not None:
            m[y <= self.min] = 0.0
        if self.max is not None:
            m[y >= self.max] = 0.0
        return -m / (t * t)

    def _betafn(self, t: ndarray) -> ndarray:
        return self._Dfn(t) * t

    def _bfn(self, t: ndarray) -> ndarray:
        D = self._Dfn(t)
        with warnings.catch_warnings(record=True) as w:
            return np.where(D == 0.0, 0.0, (self._Dfn2(t) - self.primary._Dfn2(t)) / (D * D))

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'c', 'Pivot point of early- and late-time functions [vol/vol]',
                0.0, None,
                lambda r, n: r.uniform(0.0, 1e6, n),
                exclude_lower_bound=True),
            ParamDesc(
                'm0', 'Early-time slope before pivot point',
                -10.0, 10.0,
                lambda r, n: r.uniform(-10.0, 10.0, n)),
            ParamDesc(
                'm', 'Late-time slope after pivot point',
                -1.0, 1.0,
                lambda r, n: r.uniform(-1.0, 1.0, n)),
            ParamDesc(
                't0', 'Time of pivot point [days]',
                0, None,
                lambda r, n: r.uniform(0.0, 1e5, n),
                exclude_lower_bound=True)
        ]



"""
Decline Curve Models
Copyright © 2020 David S. Fulford

Author
------
David S. Fulford
Derrick W. Turk

Notes
-----
Created on August 5, 2019
"""

from math import exp, log, log1p, ceil as ceiling, floor
import warnings

import dataclasses as dc
from dataclasses import dataclass

from numpy import ndarray
import numpy as np

from scipy.special import expi as ei, gammainc  # type: ignore
from scipy.integrate import fixed_quad  # type: ignore

from abc import ABC, abstractmethod
from typing import (TypeVar, Type, List, Dict, Tuple, Any,
                    Sequence, Optional, Callable, ClassVar, Union)
from typing import cast

from .base import (ParamDesc, DeclineCurve, PrimaryPhase, SecondaryPhase,
                   DAYS_PER_MONTH, DAYS_PER_YEAR)


@dataclass
class NullPrimaryPhase(PrimaryPhase):
    """
    A null `PrimaryPhase` class that always returns zeroes.

    Parameters
    ----------
      None
    """

    def _set_defaults(self) -> None:
        # Do not associate with the null secondary phase
        pass

    def _qfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Dfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _Dfn2(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _betafn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    def _bfn(self, t: ndarray) -> ndarray:
        return np.zeros_like(t, dtype=float)

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return []


class MultisegmentHyperbolic(PrimaryPhase):
    """
    A base class for Hyperbolic Models that generalizes for any representation of
    hyperbolic "Arps'-type" models. Each child class must implement the `_segments`
    function which generates the initial parameters of an arbitary number of
    hyperbolic segments.
    """

    T_IDX: ClassVar[int] = 0
    Q_IDX: ClassVar[int] = 1
    D_IDX: ClassVar[int] = 2
    B_IDX: ClassVar[int] = 3
    N_IDX: ClassVar[int] = 4

    segment_params: ndarray

    @abstractmethod
    def _segments(self) -> ndarray:
        """
        Precache the initial conditions of each hyperbolic segment. Should assign a list of params
        for the start condition of each segment like:

        self.params = params = np.array([
            [t_1, q_1, D_1, b_1, N_1],
            [t_2, q_2, D_2, b_2, N_2],
            [..., ..., ..., ..., ...],
            [t_m, q_n, D_n, b_n, N_m],
        ], dtype=np.float)
        """
        raise NotImplementedError

    def _validate(self) -> None:
        # this is a little naughty: bypass the "frozen" protection, just this once...
        # naturally, this should only be called during the __post_init__ process
        object.__setattr__(self, 'segment_params', self._segments())

    @staticmethod
    def _qcheck(t0: float, q: float, D: float, b: float, N: float,
                t: Union[float, ndarray]) -> ndarray:
        """
        Compute the proper Arps form of q
        """
        dt = DeclineCurve._validate_ndarray(t - t0)

        if D == 0.0:
            return np.full_like(t, q, dtype=float)

        if b < 1e-10:
            with warnings.catch_warnings(record=True) as w:
                return q * np.exp(-D * dt)

        with warnings.catch_warnings(record=True) as w:
            return q / (1.0 + D * b * dt) ** (1.0 / b)

    @staticmethod
    def _Ncheck(t0: float, q: float, D: float, b: float, N: float,
                t: Union[float, ndarray]) -> ndarray:
        """
        Compute the proper Arps form of N
        """
        dt = DeclineCurve._validate_ndarray(t - t0)

        if q <= 0.0:
            return np.atleast_1d(N) + np.zeros_like(t, dtype=float)

        if D <= 0.0:
            return np.atleast_1d(N + q * dt)

        if b <= 1e-6:
            with warnings.catch_warnings(record=True) as w:
                return N + -q / D * np.expm1(-D * dt)

        if abs(1.0 - b) == 0.0:
            return N + q / D * np.log1p(D * dt)

        with warnings.catch_warnings(record=True) as w:
            return N + q / ((1.0 - b) * D) * (1.0 - (1.0 + b * D * dt) ** (1.0 - 1.0 / b))

    @staticmethod
    def _Dcheck(t0: float, q: float, D: float, b: float, N: float,
                t: Union[float, ndarray]) -> ndarray:
        """
        Compute the proper Arps form of D
        """
        dt = DeclineCurve._validate_ndarray(t - t0)

        if D == 0.0:
            return np.full_like(t, D, dtype=float)

        return D / (1.0 + D * b * dt)

    @staticmethod
    def _Dcheck2(t0: float, q: float, D: float, b: float, N: float,
                 t: Union[float, ndarray]) -> ndarray:
        """
        Compute the derivative of the proper Arps form of D
        """
        dt = DeclineCurve._validate_ndarray(t - t0)

        if D == 0.0:
            return np.full_like(t, D, dtype=float)

        Denom = 1.0 + D * b * dt
        return -b * D * D / (Denom * Denom)

    def _vectorize(self, fn: Callable[..., ndarray], t: Union[float, ndarray]) -> ndarray:
        """
        Vectorize the computation of a parameter
        """
        t = cast(ndarray, np.atleast_1d(t))
        p = self.segment_params
        x = np.zeros_like(t, dtype=np.float)

        for i in range(p.shape[0]):
            where_seg = t >= p[i, self.T_IDX]
            if i < p.shape[0] - 1:
                where_seg = where_seg & (t < p[i + 1, self.T_IDX])

            x[where_seg] = fn(*p[i], t[where_seg])

        return x

    def _qfn(self, t: ndarray) -> ndarray:
        return self._vectorize(self._qcheck, t)

    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        return self._vectorize(self._Ncheck, t)

    def _Dfn(self, t: ndarray) -> ndarray:
        return self._vectorize(self._Dcheck, t)

    def _Dfn2(self, t: ndarray) -> ndarray:
        return self._vectorize(self._Dcheck2, t)

    def _betafn(self, t: ndarray) -> ndarray:
        return self._vectorize(self._Dcheck, t) * t

    def _bfn(self, t: ndarray) -> ndarray:
        return self._vectorize(lambda *p: p[self.B_IDX], t)

    @classmethod
    def nominal_from_secant(cls, D: float, b: float) -> float:
        with warnings.catch_warnings(record=True) as w:
            if b != 0:
                return ((1.0 - D) ** -b - 1.0) / b
            else:
                return cls.nominal_from_tangent(D)

    @classmethod
    def secant_from_nominal(cls, D: float, b: float) -> float:
        with warnings.catch_warnings(record=True) as w:
            if b != 0:
                return 1.0 - 1.0 / (1.0 + D * b) ** (1.0 / b)
            else:
                return cls.tangent_from_nominal(D)

    @classmethod
    def nominal_from_tangent(cls, D: float) -> float:
        return -log1p(-D)

    @classmethod
    def tangent_from_nominal(cls, D: float) -> float:
        return 1.0 - exp(-D)


@dataclass(frozen=True)
class MH(MultisegmentHyperbolic):
    """
    Modified Hyperbolic Model

    Robertson, S. 1988. Generalized Hyperbolic Equation.
    Available from SPE, Richardson, Texas, USA. SPE-18731-MS.

    Parameters
    ----------
      qi: float
        The initial production rate in units of ``volume / day``.

      Di: float
        The initial decline rate in secant effective decline aka annual
        effective percent decline, i.e.
        ``Di = 1 - q(t=1 year) / qi``, or
        ``Di = 1 - (1 + Dnom * b * 365.25) ** (-1 / b)``

        where ``Dnom`` is defined as ``d[ln q] / dt`` and has units of ``1 / day``.

      bi: float
        The (initial) hyperbolic parameter, defined as ``d[1 / D] / dt``.
        This parameter is dimensionless.

      Dterm: float
        The terminal secant effective decline rate aka annual effective percent decline.
    """
    qi: float
    Di: float
    bi: float
    Dterm: float

    def _validate(self) -> None:
        if self.nominal_from_secant(self.Di, self.bi) < self.nominal_from_tangent(self.Dterm):
            raise ValueError('Di < Dterm')
        super()._validate()

    def _segments(self) -> ndarray:
        """
        Precache the initial conditions of each hyperbolic segment.
        """
        Di_nom = self.nominal_from_secant(self.Di, self.bi) / DAYS_PER_YEAR
        Dterm_nom = self.nominal_from_tangent(self.Dterm) / DAYS_PER_YEAR

        if Di_nom <= 0.0 or Dterm_nom <= 0.0 or self.bi == 0.0:
            return np.array([
                [0.0, self.qi, Di_nom, self.bi, 0.0]
            ], dtype=float)

        tterm = ((1.0 / Dterm_nom) - (1.0 / Di_nom)) / self.bi
        qterm = self._qcheck(0.0, self.qi, Di_nom, self.bi, 0.0, np.array(tterm))
        Nterm = self._Ncheck(0.0, self.qi, Di_nom, self.bi, 0.0, np.array(tterm))

        return np.array([
            [0.0, self.qi, Di_nom, self.bi, 0.0],
            [tterm, qterm, Dterm_nom, 0.0, Nterm]
        ], dtype=float)

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'qi', 'Initial rate [vol/day]',
                0.0, None,
                lambda r, n: r.uniform(1e-10, 1e6, n)),
            ParamDesc(  # TODO
                'Di', 'Initial decline [sec. eff. / yr]',
                0.0, 1.0,
                lambda r, n: r.uniform(0.0, 1.0, n),
                exclude_upper_bound=True),
            ParamDesc(
                'bi', 'Hyperbolic exponent',
                0.0, 2.0,
                lambda r, n: r.uniform(0.0, 2.0, n)),
            ParamDesc(  # TODO
                'Dterm', 'Terminal decline [tan. eff. / yr]',
                0.0, 1.0,
                lambda r, n: np.zeros(n, dtype=float),
                exclude_upper_bound=True)
        ]


@dataclass(frozen=True)
class THM(MultisegmentHyperbolic):
    """
    Transient Hyperbolic Model

    Fulford, D. S., and Blasingame, T. A. 2013. Evaluation of Time-Rate
    Performance of Shale Wells using the Transient Hyperbolic Relation.
    Presented at SPE Unconventional Resources Conference – Canada in Calgary,
    Alberta, Canda, 5–7 November. SPE-167242-MS.
    https://doi.org/10.2118/167242-MS.


    Analytic Approximation

    Fulford, D.S. 2018. A Model-Based Diagnostic Workflow for Time-Rate
    Performance of Unconventional Wells. Presented at Unconventional Resources
    Conference in Houston, Texas, USA, 23–25 July. URTeC-2903036.
    https://doi.org/10.15530/urtec-2018-2903036.

    Parameters
    ----------
      qi: float
        The initial production rate in units of ``volume / day``.

      Di: float
        The initial decline rate in secant effective decline aka annual
        effective percent decline, i.e.

        .. math::

            Di = 1 - \\frac{q(t=1 \, year)}{qi}

        .. math::

            Di = 1 - \\frac{(1 + 365.25 \, D_{nom} \, b)}{1 / b}

        where ``Dnom`` is defined as ``d[ln q] / dt`` and has units of ``1 / day``.

      bi: float
        The initial hyperbolic parameter, defined as ``d[1 / D] / dt``. Is dimensionless.
        Advised to always be set to ``2.0`` to represent transient linear flow.
        See literature for more details.

      bi: float
        The final hyperbolic parameter after transition. Represents the boundary-dominated or
        boundary-influenced flow regime.

      telf: float
        The time to end of linear flow in units of ``day``, or more specifically the time at
        which ``b(t) < bi``. Visual end of half slope occurs ``~2.5x`` after ``telf``.

      bterm: Optional[float] = None
        The terminal value of the hyperbolic parameter. Has two interpretations:

        If ``tterm > 0`` then the terminal regime is a hyperbolic regime with ``b = bterm``
        and the parameter is given as the hyperbolic parameter.

        If ``tterm = 0`` then the terminal regime is an exponential regime with ``Dterm = bterm``
        and the parameter is given as secant effective decline.

      tterm: Optional[float] = None
        The time to start of the terminal regime. Setting ``tterm = 0.0`` creates an exponential
        terminal regime, while setting ``tterm > 0.0`` creates a hyperbolic terminal regime.
    """
    qi: float
    Di: float
    bi: float
    bf: float
    telf: float
    bterm: float = 0.0
    tterm: float = 0.0

    EXP_GAMMA: ClassVar[float] = exp(0.5572156)
    EXP_1: ClassVar[float] = exp(1.0)

    def _validate(self) -> None:
        # TODO: do we want to deal with optional params at all?
        # if (self.bterm is None and tterm is not None) or (bterm is not None and tterm is None):
        #     raise ValueError('Must supply both or none of tterm and bterm.')
        if self.bi < self.bf:
            raise ValueError('bi < bf')
        if self.bf < self.bterm and self.tterm != 0.0:
            # raise ValueError('bf < bterm and tterm != 0')
            # cheat to fix this
            # object.__setattr__(self, 'bterm', self.bf)
            pass
        if self.tterm != 0.0 and self.tterm * DAYS_PER_YEAR < self.telf:
            raise ValueError('tterm < telf')
        super()._validate()

    def _segments(self) -> ndarray:

        t1 = 0.0
        t2 = self.telf * (self.EXP_1 - 1.0)
        t3 = self.telf * (self.EXP_1 + 1.0)
        tterm = self.tterm * DAYS_PER_YEAR

        b1 = self.bi
        b2 = self.bi - ((self.bi - self.bf) / self.EXP_1)
        b3 = self.bf
        bterm = self.bterm

        q1 = self.qi
        D1 = self.nominal_from_secant(self.Di, self.bi) / DAYS_PER_YEAR
        N1 = 0.0

        if tterm == 0.0 and bterm == 0.0:
            # no terminal segment
            segments = np.array(
                [
                    [t1, q1, D1, b1, N1],
                    [t2, None, None, b2, None],
                    [t3, None, None, b3, None]
                ],
                dtype=np.float,
            )

        elif tterm != 0.0:
            # hyperbolic terminal segment
            t4 = tterm if tterm >= t3 else self.telf * 7.0
            b4 = min(bterm, b3)
            segments = np.array(
                [
                    [t1, q1, D1, b1, N1],
                    [t2, None, None, b2, None],
                    [t3, None, None, b3, None],
                    [t4, None, None, b4, None],
                ],
                dtype=np.float,
            )

        elif tterm == 0.0 and bterm != 0.0:
            # exponential terminal segment
            D2 = self._Dcheck(t1, q1, D1, b1, 0.0, t2).item()
            q2 = self._qcheck(t1, q1, D1, b1, 0.0, t2).item()
            D3 = self._Dcheck(t2, q2, D2, b2, 0.0, t3).item()
            D4 = self.nominal_from_tangent(bterm) / DAYS_PER_YEAR
            b4 = 0.0
            if b3 <= 0:
                t4 = t3
            else:
                t4 = max(t3, t3 + (1.0 / D4 - 1.0 / D3) / b3)

            if t4 == t3:
                segments = np.array(
                    [
                        [t1, q1, D1, b1, N1],
                        [t2, None, None, b2, None],
                        [t4, None, None, b4, None],
                    ],
                    dtype=np.float,
                )
            else:
                segments = np.array(
                    [
                        [t1, q1, D1, b1, N1],
                        [t2, None, None, b2, None],
                        [t3, None, None, b3, None],
                        [t4, None, None, b4, None],
                    ],
                    dtype=np.float,
                )

        # Compute initial values for each segment after the first, from the
        #   previous segment's values
        for i in range(segments.shape[0] - 1):
            p = [*segments[i], segments[i + 1, self.T_IDX]]
            segments[i + 1, self.D_IDX] = self._Dcheck(*p)
            segments[i + 1, self.Q_IDX] = self._qcheck(*p)
            segments[i + 1, self.N_IDX] = self._Ncheck(*p)

        return segments

    def transient_rate(self, t: Union[float, ndarray], **kwargs: Any) -> ndarray:
        """
        Compute the rate function using full definition.
        Uses :func:`scipy.integrate.fixed_quad` to integrate :func:`transient_D`.

        .. math::

            q(t) = e^{-\int_0^t D(t) \, dt}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of time values to evaluate.

          **kwargs
            Additional keyword arguments passed to :func:`scipy.integrate.fixed_quad`.

        Returns
        -------
          numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._transqfn(t, **kwargs)

    def transient_cum(self, t: Union[float, ndarray], **kwargs: Any) -> ndarray:
        """
        Compute the cumulative volume function using full definition.
        Uses :func:`scipy.integrate.fixed_quad` to integrate :func:`transient_q`.

        .. math::

            N(t) = \int_0^t q(t) \, dt

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of time values to evaluate.

          **kwargs
            Additional keyword arguments passed to :func:`scipy.integrate.fixed_quad`.

        Returns
        -------
          numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._transNfn(t, **kwargs)

    def transient_D(self, t: Union[float, ndarray]) -> ndarray:
        """
        Compute the D-parameter function using full definition.

        .. math::

            D(t) = \\frac{1}{\\frac{1}{Di} + b_i t + \\frac{bi - bf}{c}
            (\\textrm{Ei}[-e^{-c \, (t -t_{elf}) + e^(\\gamma)}]
            - \\textrm{Ei}[-e^{c \, t_{elf} + e^(\\gamma)}])}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of time values to evaluate.

        Returns
        -------
          numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._transDfn(t)

    def transient_beta(self, t: Union[float, ndarray]) -> ndarray:
        """
        Compute the beta-parameter function using full definition.

        .. math::

            \\beta(t) = \\frac{t}{\\frac{1}{Di} + b_i t + \\frac{bi - bf}{c}
            (\\textrm{Ei}[-e^{-c \, (t -t_{elf}) + e^(\\gamma)}]
            - \\textrm{Ei}[-e^{c \, t_{elf} + e^(\\gamma)}])}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of time values to evaluate.

        Returns
        -------
          numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._transDfn(t) * t

    def transient_b(self, t: Union[float, ndarray]) -> ndarray:
        """
        Compute the b-parameter function using full definition.

        .. math::

            b(t) = b_i - (b_i - b_f) e^{-\\textrm{exp}[{-c * (t - t_{elf}) + e^{\\gamma}}]}

        where:

        .. math::

            c & = \\frac{e^{\\gamma}}{1.5 \, t_{elf}} \\\\
            \\gamma & = 0.57721566... \; \\textrm{(Euler-Mascheroni constant)}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of time values to evaluate.

        Returns
        -------
          numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._transbfn(t)

    def _transNfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        with warnings.catch_warnings(record=True) as w:
            return self._integrate_with(self._transqfn, t, **kwargs)

    def _transqfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        qi = self.qi
        Dnom_i = self.nominal_from_secant(self.Di, self.bi) / DAYS_PER_YEAR
        with warnings.catch_warnings(record=True) as w:
            return qi * np.exp(Dnom_i - self._integrate_with(self._transDfn, t, **kwargs))

    def _transDfn(self, t: ndarray) -> ndarray:

        t = cast(ndarray, np.atleast_1d(t))
        qi = self.qi
        bi = self.bi
        bf = self.bf
        telf = self.telf
        bterm = self.bterm
        tterm = self.tterm * DAYS_PER_YEAR

        if self.Di == 0.0:
            return np.full_like(t, 0.0, dtype=float)

        Dnom_i = self.nominal_from_secant(self.Di, self.bi) / DAYS_PER_YEAR

        if Dnom_i <= 0.0:
            assert 'unreachable: negative Dnom in _transDfn'  # pragma: no cover

        if telf > 0.001:
            # transient function
            c = self.EXP_GAMMA / (1.5 * telf)
            with warnings.catch_warnings(record=True) as w:
                D = 1.0 / (
                    1.0 / Dnom_i
                    + bi * t
                    + (bi - bf) / c * ei(-np.exp(-c * (t - telf) + self.EXP_GAMMA))
                    - ei(-np.exp(c * telf + self.EXP_GAMMA))
                )

        else:
            # telf is too small to compute transient function
            D = self._Dcheck(0.0, qi, Dnom_i, bf, 0.0, t)
            Dterm = self._Dcheck(0.0, qi, Dnom_i, bf, 0.0, tterm).item()

        # terminal regime
        if tterm != 0.0 or bterm != 0:
            if tterm > 0.0:
                # hyperbolic
                where_term = t > tterm
                if np.count_nonzero(where_term) > 0:
                    Dterm = D[where_term][-1].item()
                else:
                    Dterm = None
            elif tterm == 0.0:
                # exponential
                Dterm = self.nominal_from_tangent(bterm) / DAYS_PER_YEAR
                where_term = Dterm > D

            if np.count_nonzero(where_term) > 0:
                D[where_term] = self._Dcheck(
                    tterm, 1.0, Dterm, bterm, 0.0, t[where_term])

        # TODO: is this needed?
        # where_nan = np.isnan(D) & np.isfinite(D)
        # if np.count_nonzero(where_nan):
        #     D[where_nan] = 0.0  # pragma: no cover

        return D

    def _transbfn(self, t: ndarray) -> ndarray:

        t = cast(ndarray, np.atleast_1d(t))
        bi = self.bi
        bf = self.bf
        telf = self.telf
        bterm = self.bterm
        tterm = self.tterm

        if telf > 0.0:
            c = self.EXP_GAMMA / (1.5 * telf)
            b = bi - (bi - bf) * np.exp(-np.exp(-c * (t - telf) + self.EXP_GAMMA))
        else:
            b = np.full_like(t, bf, dtype=float)

        # terminal regime
        if tterm != 0.0 or bterm != 0:
            if tterm > 0.0:
                # hyperbolic
                where_term = t > tterm
                _bterm = bterm
            elif tterm == 0.0:
                # exponential
                Dterm = self.nominal_from_tangent(bterm) / DAYS_PER_YEAR
                D = self._transDfn(t)
                where_term = Dterm > D
                _bterm = 0.0

            b[where_term] = _bterm

        return b

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'qi', 'Initial rate [vol/day]',
                0.0, None,
                lambda r, n: r.uniform(1.0, 2e4, n)),
            ParamDesc(  # TODO
                'Di', 'Initial decline [sec. eff. / yr]',
                0.0, 1.0,
                lambda r, n: r.uniform(0.0, 1.0, n),
                exclude_upper_bound=True),
            ParamDesc(
                'bi', 'Initial hyperbolic exponent',
                0.0, 2.0,
                lambda r, n: np.full(n, 2.0)),
            ParamDesc(  # TODO
                'bf', 'Final hyperbolic exponent',
                0.0, 2.0,
                lambda r, n: r.uniform(0.0, 1.0, n)),
            ParamDesc(  # TODO
                'telf', 'Time to end of linear flow [days]',
                None, None,
                lambda r, n: r.uniform(1e-10, 365.25, n)),
            ParamDesc(
                'bterm', 'Terminal hyperbolic exponent',
                0.0, 2.0,
                lambda r, n: np.full(n, 0.0)),
            ParamDesc(
                'tterm', 'Terminal time [years]',
                0.0, None,
                lambda r, n: np.full(n, 0.0))
        ]


@dataclass(frozen=True)
class PLE(PrimaryPhase):
    """
    Power-Law Exponential Model

    Ilk, D., Perego, A. D., Rushing, J. A., and Blasingame, T. A. 2008.
    Exponential vs. Hyperbolic Decline in Tight Gas Sands – Understanding
    the Origin and Implications for Reserve Estimates Using Arps Decline Curves.
    Presented at SPE Annual Technical Conference and Exhibition in Denver,
    Colorado, USA, 21–24 September. SPE-116731-MS. https://doi.org/10.2118/116731-MS.

    Ilk, D., Rushing, J. A., and Blasingame, T. A. 2009.
    Decline Curve Analysis for HP/HT Gas Wells: Theory and Applications.
    Presented at SPE Annual Technical Conference and Exhibition in New Orleands,
    Louisiana, USA, 4–7 October. SPE-125031-MS. https://doi.org/10.2118/125031-MS.

    Parameters
    ----------
      qi: float
        The initial production rate in units of ``volume / day``.

      Di: float
        The initial decline rate in nominal decline rate defined as ``d[ln q] / dt``
        and has units of ``1 / day``.

      Dterm: float
        The terminal decline rate in nominal decline rate, has units of ``1 / day``.

      n: float
        The n exponent.
    """
    qi: float
    Di: float
    Dinf: float
    n: float

    def _validate(self) -> None:
        if self.Dinf > self.Di:
            raise ValueError('Dinf > Di')

    def _qfn(self, t: ndarray) -> ndarray:
        qi = self.qi
        Di = self.Di
        Dinf = self.Dinf
        n = self.n
        return qi * np.exp(-Di * t ** n - Dinf * t)

    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        with warnings.catch_warnings(record=True) as w:
            return self._integrate_with(self._qfn, t, **kwargs)

    def _Dfn(self, t: ndarray) -> ndarray:
        Di = self.Di
        Dinf = self.Dinf
        n = self.n
        return Dinf + Di * n * t ** (n - 1.0)

    def _Dfn2(self, t: ndarray) -> ndarray:
        Di = self.Di
        Dinf = self.Dinf
        n = self.n
        return Dinf + Di * n * (n - 1.0) * t ** (n - 2.0)

    def _betafn(self, t: ndarray) -> ndarray:
        Di = self.Di
        Dinf = self.Dinf
        n = self.n
        return Dinf * t + Di * n * t ** n

    def _bfn(self, t: ndarray) -> ndarray:
        Di = self.Di
        Dinf = self.Dinf
        n = self.n
        Denom = (Dinf * t + Di * n * t ** n)
        return Di * (1.0 - n) * n * t ** n / (Denom * Denom)

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'qi', 'Initial rate [vol/day]',
                0, None,
                lambda r, n: r.uniform(1e-10, 1e6, n)),
            ParamDesc(
                'Di', 'Initial decline rate [/day]',
                0.0, None,
                lambda r, n: r.uniform(0.0, 1e3, n)),
            ParamDesc(
                'Dinf', 'Terminal decline rate [/day]',
                0, None,
                lambda r, n: r.uniform(0.0, 1e3, n)),
            ParamDesc(
                'n', 'PLE exponent',
                0.0, 1.0,
                lambda r, n: r.uniform(1e-6, 1.0, n),
                exclude_lower_bound=True,
                exclude_upper_bound=True),
        ]


@dataclass(frozen=True)
class SE(PrimaryPhase):
    """
    Stretched Exponential

    Valkó, P. P. Assigning Value to Stimulation in the Barnett Shale:
    A Simultaneous Analysis of 7000 Plus Production Histories and Well
    Completion Records. 2009. Presented at SPE Hydraulic Fracturing
    Technology Conference in College Station, Texas, USA, 19–21 January.
    SPE-119369-MS. https://doi.org/10.2118/119369-MS.

    Parameters
    ----------
      qi: float
        The initial production rate in units of ``volume / day``.

      tau: float
        The tau parameter in units of ``day ** n``. Equivalent to:

        .. math::

            \\tau = D^n

      n: float
        The ``n`` exponent.
    """
    qi: float
    tau: float
    n: float

    def _qfn(self, t: ndarray) -> ndarray:
        qi = self.qi
        tau = self.tau
        n = self.n
        return qi * np.exp(-(t / tau) ** n)

    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        qi = self.qi
        tau = self.tau
        n = self.n
        return qi * tau / n * gammainc(1.0 / n, (t / tau) ** n)

    def _Dfn(self, t: ndarray) -> ndarray:
        tau = self.tau
        n = self.n
        return n * tau ** -n * t ** (n - 1.0)

    def _Dfn2(self, t: ndarray) -> ndarray:
        tau = self.tau
        n = self.n
        return n * (n - 1.0) * tau ** -n * t ** (n - 2.0)

    def _betafn(self, t: ndarray) -> ndarray:
        tau = self.tau
        n = self.n
        return n * tau ** -n * t ** n

    def _bfn(self, t: ndarray) -> ndarray:
        tau = self.tau
        n = self.n
        return (1.0 - n) / n * tau ** n * t ** -n

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'qi', 'Initial rate [vol/day]',
                0.0, None,
                lambda r, n: r.uniform(1e-10, 1e6, n)),
            ParamDesc(
                'tau', 'tau',
                1e-10, 1e4,
                lambda r, n: r.uniform(1e-10, 1e4, n)),
            ParamDesc(
                'n', 'SE exponent',
                1e-10, 1.0,
                lambda r, n: r.uniform(1e-10, 1.0, n),
                exclude_upper_bound=True),
        ]


@dataclass(frozen=True)
class Duong(PrimaryPhase):
    """
    Duong Model

    Duong, A. N. 2001. Rate-Decline Analysis for Fracture-Dominated
    Shale Reservoirs. SPE Res Eval & Eng 14 (3): 377–387. SPE-137748-PA.
    https://doi.org/10.2118/137748-PA.

    Parameters
    ----------
      qi: float
        The initial production rate in units of ``volume / day`` *defined at ``t=1 day``*.

      a: float
        The ``a`` parameter. Roughly speaking, controls slope of the :func:``q(t)`` function.

      m: float
        The ``m`` parameter. Roughly speaking, controls curvature of the :func:``q(t)`` function.
    """
    qi: float
    a: float
    m: float

    def _qfn(self, t: ndarray) -> ndarray:
        qi = self.qi
        a = self.a
        m = self.m
        with warnings.catch_warnings(record=True) as w:
            q = np.where(t == 0.0, 0.0,
                         qi * t ** -m * np.exp(a / (1.0 - m) * (t ** (1.0 - m) - 1.0)))
        return q

    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        qi = self.qi
        a = self.a
        m = self.m
        with warnings.catch_warnings(record=True) as w:
            N = np.where(t == 0.0, 0.0, qi / a * np.exp(a / (1.0 - m) * (t ** (1.0 - m) - 1.0)))
        return N

    def _Dfn(self, t: ndarray) -> ndarray:
        a = self.a
        m = self.m
        # alternative form: D = m * t ** -1.0 - a * t ** -m
        return m / t - a * t ** -m

    def _Dfn2(self, t: ndarray) -> ndarray:
        a = self.a
        m = self.m
        # alternative form: D = m * t ** -1.0 - a * t ** -m
        return -m / (t * t) + m * a * t ** (-m - 1.0)

    def _betafn(self, t: ndarray) -> ndarray:
        a = self.a
        m = self.m
        return m - a * t ** (1.0 - m)

    def _bfn(self, t: ndarray) -> ndarray:
        a = self.a
        m = self.m
        Denom = a * t - m * t ** m
        with warnings.catch_warnings(record=True) as w:
            return np.where(
                Denom == 0.0, 0.0, m * t ** m * (t ** m - a * t) / (Denom * Denom))

    @classmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        return [
            ParamDesc(
                'qi', 'Initial rate [vol/day]',
                0.0, None,
                lambda r, n: r.uniform(1.0, 2e4, n)),
            ParamDesc(
                'a', 'a',
                1.0, None,
                lambda r, n: r.uniform(1.0, 10.0, n)),
            ParamDesc(
                'm', 'm',
                1.0, None,
                lambda r, n: r.uniform(1.0, 10.0, n),
                exclude_lower_bound=True)
        ]



"""
Decline Curve Models
Copyright © 2020 David S. Fulford

Author
------
David S. Fulford
Derrick W. Turk

Notes
-----
Created on August 5, 2019
"""

from math import exp, log, log1p, ceil as ceiling, floor
from itertools import starmap
import warnings

import dataclasses as dc
from dataclasses import dataclass

from numpy import ndarray
import numpy as np
from numpy.random import RandomState

from scipy.special import expi as ei, gammainc  # type: ignore
from scipy.integrate import fixed_quad  # type: ignore

from abc import ABC, abstractmethod
from typing import (TypeVar, Type, List, Dict, Tuple, Any,
                    Sequence, Iterator, Optional, Callable, ClassVar, Union)
from typing import cast


DAYS_PER_MONTH = 365.25 / 12.0
DAYS_PER_YEAR = 365.25


@dataclass(frozen=True)
class ParamDesc():
    name: str
    description: str
    lower_bound: Optional[float]
    upper_bound: Optional[float]
    naive_gen: Callable[[RandomState, int], ndarray]
    exclude_lower_bound: bool = False
    exclude_upper_bound: bool = False


def get_time(start: float = 1.0, end: float = 1e5, n: int = 101) -> ndarray:
    """
    Get a time array to evaluate with.

    Parameters
    ----------
      start: float
        The first time value of the array.

      end: float
        The last time value of the array.

      n: int
        The number of element in the array.

    Returns
    -------
      time: numpy.ndarray[float]
        An evenly-logspaced time series.
    """
    return 10.0 ** np.linspace(log(start), log(end), n)


def get_time_monthly_vol(start: float = 1, end: int = 10_000) -> ndarray:
    """
    Get a time array to evaluate with.

    Parameters
    ----------
      start: float
        The first time value of the array.

      end: float
        The last time value of the array.

    Returns
    -------
      time: numpy.ndarray[float]
        An evenly-monthly-spaced time series
    """
    return (np.arange(start, end // DAYS_PER_MONTH) + 1) * DAYS_PER_MONTH


class DeclineCurve(ABC):
    """
    Base class for decline curve models. Each model must implement the defined
    abstract methods.
    """

    def rate(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model rate function:

        .. math::

            q(t) = f(t)

        where ``f(t)`` is defined by each model.

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          rate: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._qfn(t)

    def cum(self, t: Union[float, ndarray], **kwargs: Any) -> ndarray:
        """
        Defines the model cumulative volume function:

        .. math::

            N(t) = \int_0^t q \, dt

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

          **kwargs
            Additional arguments passed to :func:`scipy.integrate.fixed_quad` if needed.

        Returns
        -------
          cumulative volume: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._Nfn(t, **kwargs)

    def interval_vol(self, t: Union[float, ndarray], t0: Optional[Union[float, ndarray]] = None,
                     **kwargs: Any) -> ndarray:
        """
        Defines the model interval volume function:

        .. math::

            N(t) = \int_{t_{i-1}}^{t_i} q \, dt

        for each element of ``t``.

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of interval end times at which to evaluate the function.

          t0: Optional[Union[float, numpy.ndarray[float]]]
            A start time of the first interval. If not given, the first element
            of ``t`` is used.

          **kwargs
            Additional arguments passed to :func:`scipy.integrate.fixed_quad` if needed.

        Returns
        -------
          interval volume: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        if t0 is None:
            t0 = t[0]
        t0 = cast(ndarray, np.atleast_1d(t0).astype(float))
        return np.diff(self._Nfn(t, **kwargs), prepend=self._Nfn(t0))

    def monthly_vol(self, t: Union[float, ndarray], t0: Optional[Union[float, ndarray]] = None,
                    **kwargs: Any) -> ndarray:
        """
        Defines the model interval volume function transformed to equivalent monthly volumes:

        .. math::

            N(t) = \int_{t-{1 \, month}}^{t} q \, dt

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of interval end times at which to evaluate the function.

          t0: Optional[Union[float, numpy.ndarray[float]]]
            A start time of the first interval. If not given, assumed to be zero.

          **kwargs
            Additional arguments passed to :func:`scipy.integrate.fixed_quad` if needed.

        Returns
        -------
          monthly equivalent volume: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        if t0 is None:
            t0 = 0.0
        t0 = cast(ndarray, np.atleast_1d(t0).astype(float))
        return np.diff(self._Nfn(t, **kwargs), prepend=self._Nfn(t0)) \
            / np.diff(t, prepend=t0) * DAYS_PER_MONTH

    def D(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model D-parameter function:

        .. math::

            D \equiv \\frac{d[\\textrm{ln} \, q]}{dt}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          D-parameter: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._Dfn(t)

    def beta(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model beta-parameter function.

        .. math::

            \\beta \equiv \\frac{\\textrm{ln} \, q}{\\textrm{ln} \, t} = t \, D(t)

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          beta-parameter: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._betafn(t)

    def b(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model b-parameter function:

        .. math::

            b \equiv \\frac{d\\frac{1}{D}}{dt}

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          b-parameter: numpy.ndarray[float]
        """
        t = self._validate_ndarray(t)
        return self._bfn(t)

    @abstractmethod
    def _qfn(self, t: ndarray) -> ndarray:
        raise NotImplementedError

    @abstractmethod
    def _Nfn(self, t: ndarray, **kwargs: Any) -> ndarray:
        raise NotImplementedError

    @abstractmethod
    def _Dfn(self, t: ndarray) -> ndarray:
        raise NotImplementedError

    @abstractmethod
    def _Dfn2(self, t: ndarray) -> ndarray:
        raise NotImplementedError

    @abstractmethod
    def _betafn(self, t: ndarray) -> ndarray:
        raise NotImplementedError

    @abstractmethod
    def _bfn(self, t: ndarray) -> ndarray:
        raise NotImplementedError

    def _integrate_with(self, fn: Callable[[ndarray], ndarray], t: ndarray,
                        **kwargs: Any) -> ndarray:
        kwargs.setdefault('n', 100)
        return np.cumsum(list(starmap(
            lambda t0, t1: fixed_quad(fn, t0, t1, **kwargs)[0],
            self._iter_t(t)
        )))

    def _validate(self) -> None:
        # this will be called by the __post_init__ hook - subclasses should
        #   do any necessary additional validation or caching here
        pass

    def __post_init__(self) -> None:
        self._set_defaults()
        for desc in self.get_param_descs():
            param = getattr(self, desc.name)
            if desc.lower_bound is not None:
                if desc.exclude_lower_bound:
                    if param <= desc.lower_bound:
                        raise ValueError(f'{desc.name} <= {desc.lower_bound}')
                else:
                    if param < desc.lower_bound:
                        raise ValueError(f'{desc.name} < {desc.lower_bound}')
            if desc.upper_bound is not None:
                if desc.exclude_upper_bound:
                    if param >= desc.upper_bound:
                        raise ValueError(f'{desc.name} >= {desc.upper_bound}')
                else:
                    if param > desc.upper_bound:
                        raise ValueError(f'{desc.name} > {desc.upper_bound}')
        self._validate()

    @abstractmethod
    def _set_defaults(self) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_param_descs(cls) -> List[ParamDesc]:
        raise NotImplementedError

    # don't call this in a loop - it's a utility for e.g. test suites
    @classmethod
    def get_param_desc(cls, name: str) -> ParamDesc:
        for p in cls.get_param_descs():
            if p.name == name:
                return p  # pragma: no cover
        raise KeyError(name)

    # only exists to satisfy mypy
    def __init__(self, *args: float) -> None:
        raise NotImplementedError

    @classmethod
    def from_params(cls: Type['DeclineCurve'], params: Sequence[float]) -> 'DeclineCurve':
        """
        Construct a model from a sequence of parameters.
        """
        if len(cls.get_param_descs()) != len(params):
            raise ValueError('Params sequence does not have required length')
        return cls(*params)

    @staticmethod
    def _validate_ndarray(x: Union[float, ndarray]) -> ndarray:
        """
        Ensure the time array is a 1d arary of floats.
        """
        return np.atleast_1d(x).astype(float)

    @staticmethod
    def _iter_t(t: ndarray) -> Iterator[Tuple[float, float]]:
        """
        Yield a tuple of time intervals.
        """
        t0 = 0.0
        for t1 in t:
            yield(t0, t1)
            t0 = t1
        return

    @staticmethod
    def get_time() -> ndarray:
        return get_time()

    @staticmethod
    def get_time_monthly_vol() -> ndarray:
        return get_time_monthly_vol()


class PrimaryPhase(DeclineCurve):
    """
    Extends :class:`DeclineCurve` for a primary phase forecast.
    Adds the capability to link a secondary (associated) phase model.
    """
    secondary: 'SecondaryPhase'

    def _set_defaults(self) -> None:
        # this is a little naughty: bypass the "frozen" protection, just this once...
        # naturally, this should only be called during the __post_init__ process
        secondary = NullSecondaryPhase()
        object.__setattr__(secondary, 'primary', self)
        object.__setattr__(self, 'secondary', secondary)

    def add_secondary(self, secondary: 'SecondaryPhase') -> None:
        """
        Attach a secondary phase model to this primary phase model.

        Parameters
        ----------
          secondary: SecondaryPhase
            A model that inherits the :class:`SecondaryPhase` class.

        Returns
        -------
          None
        """
        # bypass the "frozen" protection to link to the secondary phase
        object.__setattr__(secondary, 'primary', self)
        object.__setattr__(self, 'secondary', secondary)


class SecondaryPhase(DeclineCurve):
    """
    Extends :class:`DeclineCurve` for a secondary (associated) phase forecast.
    Adds the capability to link a primary phase model.
    Defines the :meth:`gor` and :meth:`cgr` functions. Each model must implement the
    defined abstract method.

    """
    primary: 'PrimaryPhase'

    def _set_defaults(self) -> None:
        # this is a little naughty: bypass the "frozen" protection, just this once...
        # naturally, this should only be called during the __post_init__ process
        primary = NullPrimaryPhase()
        object.__setattr__(primary, 'secondary', self)
        object.__setattr__(self, 'primary', primary)

    def gor(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model GOR function.
        Implementation is idential to CGR function.

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          GOR: numpy.ndarray[float]
            The gas-oil ratio function in units of ``Bbl / scf``.
        """
        t = self._validate_ndarray(t)
        return self._yieldfn(t)

    def cgr(self, t: Union[float, ndarray]) -> ndarray:
        """
        Defines the model CGR function.
        Implementation is identical to GOR function.

        Parameters
        ----------
          t: Union[float, numpy.ndarray[float]]
            An array of times at which to evaluate the function.

        Returns
        -------
          CGR: numpy.ndarray[float]
            The condensate-gas ratio in units of ``MMscf / Bbl``.
        """
        t = self._validate_ndarray(t)
        return self._yieldfn(t)

    @abstractmethod
    def _yieldfn(self, t: ndarray) -> ndarray:
        raise NotImplementedError


# Must import these here to avoid circular dependency
from .primary import NullPrimaryPhase
from .secondary import NullSecondaryPhase



"""
Decline Curve Models
Unit Testing
Copyright © 2020 David S. Fulford

Author
------
David S. Fulford
Derrick W. Turk

Notes
-----
Created on August 5, 2019
"""
import sys
import warnings
import pytest # type: ignore
import hypothesis
from hypothesis import assume, given, settings, note, strategies as st
from typing import Any, Type, TypeVar

from math import isnan
import numpy as np

from petbox import dca
from data import rate as q_data, time as t_data

# ignores = [
#     'overflow encountered in multiply',
#     'overflow encountered in true_divide',
#     'overflow encountered in power',
#     'invalid value encountered in subtract',
#     'invalid value encountered in greater',
# ]
# for ig in ignores:
#     warnings.filterwarnings('ignore', ig)

# def signif(x, p):
#     x = np.asarray(x)
#     x_positive = np.where(np.isfinite(x) & (x != 0), np.abs(x), 10**(p-1))
#     mags = 10 ** (p - 1 - np.floor(np.log10(x_positive)))
#     return np.round(x * mags) / mags


def is_float_array_like(arr: Any, like: np.ndarray) -> bool:
    return (
        isinstance(arr, np.ndarray)
        and arr.dtype == np.float64
        and arr.shape == like.shape
    )


def is_monotonic_nonincreasing(arr: np.ndarray) -> bool:
    # a = np.diff(signif(arr, 6))
    a = np.diff(arr, 6)
    return np.all(a <= 0.0)


def is_monotonic_increasing(arr: np.ndarray) -> bool:
    # a = np.diff(signif(arr, 6))
    a = np.diff(arr, 6)
    return np.all(a > 0.0)


def is_monotonic_nondecreasing(arr: np.ndarray) -> bool:
    # a = np.diff(signif(arr, 6))
    a = np.diff(arr, 6)
    return np.all(a >= 0.0)


T = TypeVar('T', bound=dca.DeclineCurve)
def model_floats(model_cls: Type[T], param: str) -> st.SearchStrategy[float]:
    p = model_cls.get_param_desc(param)
    return st.floats(p.lower_bound, p.upper_bound,
                     exclude_min=p.exclude_lower_bound, exclude_max=p.exclude_upper_bound)


def check_model(model: dca.DeclineCurve, qi: float) -> bool:
    t = model.get_time()
    # assert is_monotonic_increasing(t)

    if isinstance(model, dca.Duong):
        t0 = 1e-3
        assert np.isclose(model.rate(np.array(1.0)), qi, atol=1e-10)
        assert np.isclose(model.cum(np.array(1.0)), qi / model.a, atol=1e-10)
    else:
        t0 = 0.0
        assert np.isclose(model.rate(np.array(0.0)), qi, atol=1e-10)
        assert np.isclose(model.cum(np.array(0.0)), 0.0, atol=1e-10)

    rate = model.rate(t)
    assert is_float_array_like(rate, t)
    # assert is_monotonic_nonincreasing(rate)
    assert np.all(np.isfinite(rate))

    cum = model.cum(t)
    assert is_float_array_like(cum, t)
    # if not isinstance(model, dca.PLE):
        # exclude PLE as it is numerically integrated
        # assert is_monotonic_nondecreasing(cum)
    assert np.all(np.isfinite(cum))

    mvolume = model.monthly_vol(t)
    mavg_rate = np.gradient(mvolume, t)
    # assert is_float_array_like(mvolume, t)
    # assert is_monotonic_nonincreasing(mavg_rate)
    assert np.all(np.isfinite(mvolume))
    assert np.all(np.isfinite(mavg_rate))

    ivolume = model.interval_vol(t)
    iavg_rate = np.gradient(ivolume, t)
    # assert is_float_array_like(ivolume, t)
    # assert is_monotonic_nonincreasing(iavg_rate)
    assert np.all(np.isfinite(ivolume))
    assert np.all(np.isfinite(iavg_rate))

    D = model.D(t)
    assert is_float_array_like(D, t)
    # assert is_monotonic_nonincreasing(D)
    assert np.all(np.isfinite(D))

    D2 = model._Dfn2(t)
    assert is_float_array_like(D2, t)
    # assert is_monotonic_nonincreasing(D2)
    assert np.all(np.isfinite(D2))

    beta = model.beta(t)
    assert is_float_array_like(beta, t)
    # TODO: what are the invariants for beta?
    D_inferred = beta / t
    # assert is_monotonic_nonincreasing(D_inferred)
    assert np.all(np.isfinite(beta))

    b = model.b(t)
    assert is_float_array_like(b, t)
    assert np.all(np.isfinite(b))

    return True


def check_yield_model(model: dca.PrimaryPhase, qi: float) -> bool:
    t = model.get_time()
    # assert is_monotonic_increasing(t)

    t0 = 0.0
    assert np.isclose(model.secondary.cum(np.array(0.0)), 0.0, atol=1e-10)

    gor = model.secondary.gor(t)
    assert is_float_array_like(gor, t)
    assert np.all(np.isfinite(gor))

    cgr = model.secondary.cgr(t)
    assert is_float_array_like(cgr, t)
    assert np.all(np.isfinite(cgr))

    rate = model.secondary.rate(t)
    assert is_float_array_like(rate, t)
    # assert is_monotonic_nonincreasing(rate)
    assert np.all(np.isfinite(rate))

    cum = model.secondary.cum(t)
    assert is_float_array_like(cum, t)
    # if not isinstance(model, dca.PLE):
        # exclude PLE as it is numerically integrated
        # assert is_monotonic_nondecreasing(cum)
    assert np.all(np.isfinite(cum))

    mvolume = model.secondary.monthly_vol(t, t0=t0)
    mavg_rate = np.gradient(mvolume, t)
    # assert is_float_array_like(mvolume, t)
    # assert is_monotonic_nonincreasing(mavg_rate)
    assert np.all(np.isfinite(mvolume))
    assert np.all(np.isfinite(mavg_rate))

    ivolume = model.secondary.interval_vol(t, t0=t0)
    iavg_rate = np.gradient(ivolume, t)
    # assert is_float_array_like(ivolume, t)
    # assert is_monotonic_nonincreasing(iavg_rate)
    assert np.all(np.isfinite(ivolume))
    assert np.all(np.isfinite(iavg_rate))

    D = model.secondary.D(t)
    assert is_float_array_like(D, t)
    # assert is_monotonic_nonincreasing(D)
    # assert np.all(np.isfinite(D))

    D2 = model.secondary._Dfn2(t)
    assert is_float_array_like(D2, t)
    # assert is_monotonic_nonincreasing(D2)
    # assert np.all(np.isfinite(D2))

    beta = model.secondary.beta(t)
    assert is_float_array_like(beta, t)
    # TODO: what are the invariants for beta?
    # D_inferred = beta / t
    # assert is_monotonic_nonincreasing(D_inferred)
    # assert np.all(np.isfinite(beta))

    b = model.secondary.b(t)
    assert is_float_array_like(b, t)
    assert np.all(np.isfinite(b))

    # der = model.secondary._derfn(np.array([0.0]))
    # NN = model.secondary._NNfn(np.array([0.0]))

    return True


def check_transient_model(model: dca.THM) -> bool:
    t = model.get_time()

    t_D = model.transient_D(t)
    assert is_float_array_like(t_D, t)
    # assert is_monotonic_nonincreasing(t_D)
    assert np.all(np.isfinite(t_D))

    t_beta = model.transient_beta(t)
    assert is_float_array_like(t_beta, t)
    # assert is_monotonic_nonincreasing(t_beta)
    assert np.all(np.isfinite(t_beta))

    t_b = model.transient_b(t)
    assert is_float_array_like(t_b, t)
    # assert is_monotonic_nonincreasing(t_b)
    assert np.all(np.isfinite(t_b))

    return True


def check_transient_model_rate_cum(model: dca.THM) -> bool:
    # these are computationally expensive, so check separately
    t = model.get_time()

    t_N = model.transient_cum(t)
    assert is_float_array_like(t_N, t)
    # assert is_monotonic_nondecreasing(t_N)
    assert np.all(np.isfinite(t_N))

    t_q = model.transient_rate(t)
    assert is_float_array_like(t_q, t)
    # assert is_monotonic_nonincreasing(t_q)
    assert np.all(np.isfinite(t_q))

    return True


def test_time_arrays():
    t = dca.get_time()
    int_t = dca.get_time_monthly_vol()

    thm = dca.THM(1000, 0.5, 2.0, 1.0, 30.0)
    t = thm.get_time()
    int_t = thm.get_time_monthly_vol()


def test_nulls():
    t = dca.get_time()
    primary = dca.NullPrimaryPhase()
    assert np.allclose(primary.rate(t), 0.0)
    assert np.allclose(primary.cum(t), 0.0)
    assert np.allclose(primary.D(t), 0.0)
    assert np.allclose(primary.beta(t), 0.0)
    assert np.allclose(primary.b(t), 0.0)
    assert np.allclose(primary._Dfn2(t), 0.0)

    secondary = dca.NullSecondaryPhase()
    assert np.allclose(secondary.gor(t), 0.0)
    assert np.allclose(secondary.cgr(t), 0.0)
    assert np.allclose(secondary.rate(t), 0.0)
    assert np.allclose(secondary.cum(t), 0.0)
    assert np.allclose(secondary.D(t), 0.0)
    assert np.allclose(secondary.beta(t), 0.0)
    assert np.allclose(secondary.b(t), 0.0)
    assert np.allclose(secondary._Dfn2(t), 0.0)


# TODO: use bounds, after we use testing to set them
@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(1e-10, 1e10),
    Dinf=st.floats(1e-10, 1e10),
    n=st.floats(1e-10, 1.0, exclude_max=True)
)
def test_PLE(qi, Di, Dinf, n):
    assume(Dinf <= Di)
    ple = dca.PLE.from_params((qi, Di, Dinf, n))
    ple = dca.PLE(qi, Di, Dinf, n)
    check_model(ple, qi)


@given(
    qi=st.floats(1e-10, 1e6),
    tau=st.floats(1e-10, 1e4),
    n=st.floats(1e-10, 1.0, exclude_max=True)
)
def test_SE(qi, tau, n):
    se = dca.SE.from_params((qi, tau, n))
    se = dca.SE(qi, tau, n)
    check_model(se, qi)


@given(
    qi=st.floats(1e-10, 1e6),
    a=st.floats(1.0, 10.0),
    m=st.floats(1.0, 10.0, exclude_min=True)
)
def test_Duong(qi, a, m):
    duong = dca.Duong.from_params((qi, a, m))
    duong = dca.Duong(qi, a, m)
    check_model(duong, qi)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(0.0, 1e6)
)
def test_THM(qi, Di, bf, telf):
    thm = dca.THM.from_params((qi, Di, 2.0, bf, telf, 0.0, 0.0))
    thm = dca.THM(qi, Di, 2.0, bf, telf, 0.0, 0.0)
    check_model(thm, qi)
    check_transient_model(thm)

    thm = dca.THM(qi, Di, 2.0, 0.0, telf)
    check_model(thm, qi)
    check_transient_model(thm)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(0.0, 1e4),
    bterm=st.floats(0.0, 1.0),
    tterm=st.floats(1e-3, 30.0),
)
def test_THM_terminal(qi, Di, bf, telf, bterm, tterm):
    assume(tterm * dca.DAYS_PER_YEAR > telf)
    assume(bterm < bf)
    thm = dca.THM(qi, Di, 2.0, bf, telf, bterm, tterm)
    check_transient_model(thm)
    check_model(thm, qi)


@given(
    qi=st.floats(1e-10, 1e6),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(0.0, 1e4),
    bterm=st.floats(0.0, 1.0),
    tterm=st.floats(5.0, 30.0),
)
def test_THM_zero_Di(qi, bf, telf, bterm, tterm):
    assume(tterm * dca.DAYS_PER_YEAR > telf)
    assume(bterm < bf)
    thm = dca.THM(qi, 0.0, 2.0, bf, telf, bterm, tterm)
    check_model(thm, qi)
    check_transient_model(thm)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    telf=st.floats(0.0, 1e4),
    bterm=st.floats(0.0, 0.5),
    tterm=st.floats(5, 30),
)
def test_THM_harmonic(qi, Di, telf, bterm, tterm):
    assume(tterm * dca.DAYS_PER_YEAR > telf)
    thm = dca.THM(qi, Di, 2.0, 1.0, telf, bterm, tterm)
    check_model(thm, qi)
    check_transient_model(thm)


def test_THM_transient_extra():
    thm = dca.THM(1000.0, 0.80, 2.0, 0.8, 30.0, 0.3, 5.0)
    check_transient_model(thm)
    check_transient_model_rate_cum(thm)

    thm = dca.THM(1000.0, 0.80, 2.0, 0.8, 30.0, 0.06, 0.0)
    check_transient_model(thm)
    check_transient_model_rate_cum(thm)

    thm = dca.THM(1000.0, 1e-10, 2.0, 0.8, 1e-5, 0.5, 0.06)
    check_transient_model(thm)
    check_transient_model_rate_cum(thm)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(0.0, 1e6),
    bterm=st.floats(1e-3, 0.3)
)
@settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
def test_THM_terminal_exp(qi, Di, bf, telf, bterm):
    assume(dca.THM.nominal_from_secant(Di, 2.0) >= dca.THM.nominal_from_tangent(bterm))
    thm = dca.THM(qi, Di, 2.0, bf, telf, bterm, 0.0)
    check_model(thm, qi)
    check_transient_model(thm)


@given(
    qi=st.floats(0.0, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bi=st.floats(0.0, 2.0),
    Dterm=st.floats(0.0, 1.0, exclude_max=True),
)
def test_MH(qi, Di, bi, Dterm):
    assume(dca.MH.nominal_from_secant(Di, bi) >= dca.MH.nominal_from_tangent(Dterm))
    mh = dca.MH(qi, Di, bi, Dterm)
    check_model(mh, qi)

    mh = dca.MH(qi, 0.0, bi, 0.0)
    check_model(mh, qi)


@given(
    qi=st.floats(0.0, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    Dterm=st.floats(0.0, 1.0, exclude_max=True),
)
def test_MH_harmonic(qi, Di, Dterm):
    assume(dca.MH.nominal_from_secant(Di, 1.0) >= dca.MH.nominal_from_tangent(Dterm))
    mh = dca.MH(qi, Di, 1.0, Dterm)
    check_model(mh, qi)


@given(
    D=st.floats(0.0, 1.0, exclude_max=True),
    b=st.floats(0.0, 2.0),
)
def test_decline_conv(D, b):
    Dnom = dca.MultisegmentHyperbolic.nominal_from_secant(D, b)
    _D = dca.MultisegmentHyperbolic.secant_from_nominal(Dnom, b)

def test_bound_errors():
    with pytest.raises(ValueError) as e:
        # < lower bound
        ple = dca.PLE(-1000, 0.8, 0.0, 0.5)

    with pytest.raises(ValueError) as e:
        # lower bound excluded
        ple = dca.PLE(1000, 0.8, 0.0, 0.0)

    with pytest.raises(ValueError) as e:
        # > upper bound
        thm = dca.THM(1000, 0.5, 2.0, 10.0, 30.0)

    with pytest.raises(ValueError) as e:
        # upper bound exluded
        thm = dca.THM(1000, 1.5, 2.0, 0.5, 30.0)

    with pytest.raises(KeyError) as e:
        # invalid parameter
        thm = dca.THM(1000, 0.5, 2.0, 0.5, 30.0)
        thm.get_param_desc('n')

    with pytest.raises(ValueError) as e:
        # invalid parameter sequence length
        thm = dca.THM.from_params([1000, 0.5, 2.0, 0.5])


def test_terminal_exceeds():
    with pytest.raises(ValueError) as e:
        # Dinf > Di
        ple = dca.PLE(1000, 0.8, 0.9, 0.5)

    with pytest.raises(ValueError) as e:
        # Dterm > Di
        mh = dca.MH(1000, 0.5, 1.0, 0.9)

    with pytest.raises(ValueError) as e:
        # bf > bi
        thm = dca.THM(1000, 0.8, 1.5, 1.6, 30.0)

    with pytest.raises(ValueError) as e:
        # tterm < telf
        thm = dca.THM(1000, 0.8, 2.0, 1.0, 200.0, 0.3, 100.0 / dca.DAYS_PER_YEAR)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(1e-10, 1e4),
    bterm=st.floats(1e-3, 0.3, exclude_max=True),
    tterm=st.floats(5.0, 30.0),
    c=st.floats(1e-10, 1e10),
    m0=st.floats(-1.0, 1.0),
    m=st.floats(-1.0, 1.0),
    t0=st.floats(1e-10, 365.25),
)
def test_yield(qi, Di, bf, telf, bterm, tterm, c, m0, m, t0):
    assume(tterm * dca.DAYS_PER_YEAR > telf)
    assume(bterm < bf)
    thm = dca.THM(qi, Di, 2.0, bf, telf, bterm, tterm)
    thm.add_secondary(dca.PLYield(c, m0, m, t0))
    check_yield_model(thm, qi)


@given(
    qi=st.floats(1e-10, 1e6),
    Di=st.floats(0.0, 1.0, exclude_max=True),
    bf=st.floats(0.0, 2.0),
    telf=st.floats(1e-10, 1e4),
    bterm=st.floats(1e-3, 0.3, exclude_max=True),
    tterm=st.floats(5.0, 30.0),
    c=st.floats(1e-10, 1e10),
    m0=st.floats(-1.0, 1.0),
    m=st.floats(-1.0, 1.0),
    t0=st.floats(1e-10, 365.25),
    _min=st.floats(0, 100.0),
    _max=st.floats(1e4, 5e5)
)
def test_yield_min_max(qi, Di, bf, telf, bterm, tterm, c, m0, m, t0, _min, _max):
    assume(tterm * dca.DAYS_PER_YEAR > telf)
    assume(bterm < bf)
    thm = dca.THM(qi, Di, 2.0, bf, telf, bterm, tterm)
    thm.add_secondary(dca.PLYield(c, m0, m, t0, _min, _max))
    check_yield_model(thm, qi)


def test_yield_errors():
    with pytest.raises(ValueError) as e:
        # < lower bound
        ple = dca.PLE(-1000, 0.8, 0.0, 0.5)

    with pytest.raises(ValueError) as e:
        # lower bound excluded
        thm = dca.PLE(1000, 0.8, 0.0, 0.0)

    with pytest.raises(ValueError) as e:
        # > upper bound
        thm = dca.THM(1000, 0.5, 2.0, 10.0, 30.0)

    with pytest.raises(ValueError) as e:
        # upper bound exluded
        thm = dca.THM(1000, 1.5, 2.0, 0.5, 30.0)

    with pytest.raises(KeyError) as e:
        # invalid parameter
        thm = dca.THM(1000, 0.5, 2.0, 0.5, 30.0)
        thm.get_param_desc('n')

    with pytest.raises(ValueError) as e:
        # invalid parameter sequence length
        thm = dca.THM.from_params([1000, 0.5, 2.0, 0.5])

@given(
    L=st.floats(0.0, 2.0),
    xlog=st.booleans(),
    ylog=st.booleans()
)
def test_bourdet(L, xlog, ylog):
    der = dca.bourdet(q_data, t_data, L, xlog, ylog)



"""
Decline Curve Models
Copyright © 2020 David S. Fulford

Author
------
David S. Fulford
Derrick W. Turk

Notes
-----
Created on August 5, 2019
"""

from math import exp, log, log1p, ceil as ceiling, floor
import warnings

import dataclasses as dc
from dataclasses import dataclass

from numpy import ndarray
import numpy as np

from scipy.special import expi as ei, gammainc  # type: ignore
from scipy.integrate import fixed_quad  # type: ignore

from abc import ABC, abstractmethod
from typing import (TypeVar, Type, List, Dict, Tuple, Any,
                    Sequence, Optional, Callable, ClassVar, Union)
from typing import cast


LOG10 = log(10)

def _get_end_L(x: ndarray, L: float, i: int = 0) -> int:
    """
    Left-end points that lay outside of distance L.
    """
    dx = x - x[i]

    k = len(dx) - 1
    idx = np.where((dx <= L) & (dx >= 0.))[0]
    if idx.size > 0:
        k = min(k, idx[-1] + 1)

    return k


def _get_end_R(x: ndarray, L: float, i: int = -1) -> int:
    """
    Right-end points that lay outside of distance L.
    """
    dx = x[i] - x

    k = 0
    idx = np.where((dx <= L) & (dx >= 0.))[0]
    if idx.size > 0:
        k = max(k, idx[0] - 1)

    return k


def _get_L(y: ndarray, x: ndarray, L: float, i: int) -> Tuple[ndarray, ndarray]:
    """
    Left-end points that lay inside of distance L.
    """
    dx = x[i] - x[:i]
    dy = y[i] - y[:i]
    idx = np.where((dx <= L) & (dx >= 0.))[0]

    if idx.size > 0:
        idx = max(0, idx[0] - 1)
        return dx[idx], dy[idx]
    else:
        return dx[-1], dy[-1]


def _get_R(y: ndarray, x: ndarray, L: float, i: int) -> Tuple[ndarray, ndarray]:
    """
    Right-end points that lay inside of distance L.
    """
    dx = x[i + 1:] - x[i]
    dy = y[i + 1:] - y[i]
    idx = np.where((dx <= L) & (dx >= 0.))[0]

    if idx.size > 0:
        idx = min(len(x) - 1, idx[-1] + 1)
        return dx[idx], dy[idx]
    else:
        return dx[0], dy[0]


def bourdet(y: ndarray, x: ndarray, L: float = 0.0,
            xlog: bool = True, ylog: bool = False) -> Tuple[ndarray, ndarray]:
    """
    Bourdet Derivative Smoothing

    Bourdet, D., Ayoub, J. A., and Pirard, Y. M. 1989. Use of Pressure Derivative in
    Well-Test Interpretation. SPE Form Eval 4 (2): 293–302. SPE-12777-PA.
    https://doi.org/10.2118/12777-PA.

    Parameters
    ----------
      y: numpy.ndarray[float]
        An array of y values to compute the derivative for.

      x: numpy.ndarray[float]
        An array of x values.

      L: float = 0.0
        Smoothing factor in units of log-cycle fractions.

      xlog: bool = True
        Calculate the derivative with respect to the log of x, i.e. ``dy / d[ln x]``.

      ylog: bool = False
        Calculate the derivative with respect to the log of y, i.e. ``d[ln y] / dx``.

    Returns
    -------
      der: numpy.ndarray[float]
        The calculated derivative.
    """

    with warnings.catch_warnings(record=True) as w:

        x = np.atleast_1d(x).astype(float)
        y = np.atleast_1d(y).astype(float)

        log_x = cast(ndarray, np.log10(x))

        if ylog:
            y = cast(ndarray, np.log(y))

        x_L = np.zeros_like(log_x, dtype=float)
        x_R = np.zeros_like(log_x, dtype=float)
        y_L = np.zeros_like(log_x, dtype=float)
        y_R = np.zeros_like(log_x, dtype=float)

        # get points for forward and backward derivatives
        k1 = _get_end_L(log_x, L)
        k2 = _get_end_R(log_x, L)

        # compute first & last points
        x_L[0] = log_x[k1] - log_x[0]
        y_L[0] = y[k1] - y[0]

        x_R[-1] = log_x[-1] - log_x[k2]
        y_R[-1] = y[-1] - y[k2]

        # compute bourdet derivative
        for i in range(k1, k2):
            x_L[i], y_L[i] = _get_L(y, log_x, L, i)
            x_R[i], y_R[i] = _get_R(y, log_x, L, i)

        # if xlog:
        x_L *= LOG10
        x_R *= LOG10

        der = (y_L / x_L * x_R + y_R / x_R * x_L) / (x_L + x_R)

        # compute forward difference at left edge
        for i in range(0, k1):
            idx = _get_end_L(log_x, L, i)

            dy = y[idx] - y[0]
            dx = log_x[idx] - log_x[0]

            # if xlog:
            dx *= LOG10

            der[i] = dy / dx

        # compute backward difference at right edge
        for i in range(k2, len(log_x)):
            idx = _get_end_R(log_x, L, i)

            dy = y[-1] - y[idx]
            dx = log_x[-1] - log_x[idx]

            # if xlog:
            dx *= LOG10

            der[i] = dy / dx

        if not xlog:
            der /= x

        return der
